import numpy as np
import scipy as sp
from ast import literal_eval
#from https://en.wikipedia.org/wiki/Just_intonation#Pythagorean_tuning
pythagorean={
    0:(1,1),
    1:(256,243),
    2:(9,8),
    3:(32,27),
    4:(81,64),
    5:(4,3),
    6:(729,512),
    7:(3,2),
    8:(128,81),
    9:(27,16),
    10:(16,9),
    11:(243,128),
    12:(2,1)
}

#Five_limit from https://en.wikipedia.org/wiki/Just_intonation#Five-limit_tuning  
just={
    0:(1,1),     # C
    1:(16,15),   # Db
    2:(9,8),     # D     
    3:(6,5),     # Eb
    4:(5,4),     # E
    5:(4,3),     # F
    6:(45,32),   # Gb
    7:(3,2),     # G
    8:(8,5),     # Ab
    9:(5,3),     # A 
    10:(9,5),    # Bb
    11:(15,8),   # B
    12:(2,1)     # C
}
frygian_tense_chromatic={
    # This is just one out of around 50 scales of According by Ptolemy https://en.wikipedia.org/wiki/Ptolemy 
    # who lived in Alexandria about the year 100 I have added some just scale notes in between
    # Ptolomy uses the distance between each note -while my system puts in the distance to the root note
    # (It took me quite a while to compile the numbers.
    # If I wanted to make all the 50 scale I should probably write a sub rootine that could use Ptolomys  )
    # Assuming that the Ptolomy Frygian tense chromatic is of type w w w h w w h, (which would be called Lydian in todays world Ptolomy's work was translated 
    # to Arabic and then to Latin, in the translation there was some confusion
    # mabye also because Ptolomy used tetrachords, when you play a schale on a lyre or a harp you start from the top and the Roman scales which start from the bottom, I don't know)
    # I have filled the  # in between semi-tones with just intervals. This is just a "strange" scale out of 50 scales.  15 years ago I compiled some 50 scales 
    # from Andrew Barkers "Greek Musical Writing" ISBN 052130220X  you can hear the scales here: http://oleviolin.com/violin/eng/PtolemaiosScales.htm
    # It took me quite som effort t
    0:(1,1),     # Parahypate mesôn (C)
    1:(16,15),   # Db  Just scales are inserted "in between the old greek"
    2:(12,11),   # Lichanos mesôn  12/11 (D)
    3:(6,5),     # Eb
    4:(14,11),   # Mese mesôn     7/6  (E)
    5:(4,3),     # F
    6:(63,44),    # Paramese dizeugmenôn 9/8 (F#)                 
    7:(56,33),   # Trité dizeugmenôn 28/27  (G)
    8:(8,5),     # Ab
    9:(98,66),   # Nete dizeugmenôn 8/7 (A)
    10:(9,5),    # Bb   
    11:(84,44),   # Paranete dizeugmenôn 9/8 (B)
    12:(2,1)     # Trite Hyperbolaion  22/21
}



def totalTime(midi_sequence,TpQ,TrackLength):  #Calculate the total time of a Midi_sequence, can handle changing tempos

    Tempos=[]
    tickNow=0
    for tone in midi_sequence:        
        if tone["function"]=="Tempo":
            tempo=int(tone["tempo"])            
            tick=tone["tick"]
            temp={"tempo":tempo,"tick":tick}            
            Tempos.append(temp)            
    if len(Tempos)==1:
        tempo=int(Tempos[0]['tempo'])
        TotalTime=tempo*TrackLength/(TpQ*1000000)          
    else:
        i=0
        TotalTime=0    
        for tempShift in Tempos:            
            if i>0:                
                lastTempo=int(Tempos[i-1]['tempo'])
                lastTick=int(Tempos[i-1]['tick'])
                tickNow=int(Tempos[i]['tick'])
                ticks=tickNow-lastTick
                TotalTime=TotalTime+lastTempo*ticks/(TpQ*1000000) 
                tempo=int(Tempos[i]['tempo'])
                print(str(float(TotalTime))+' tick' +str(tickNow))
            i=i+1
        #Last time add remaining time
        TotalTime=TotalTime+(TrackLength-tickNow)*tempo/(TpQ*1000000)          
    return TotalTime  #in seconds
def factor_2_cent(factor):
    cent=1200* np.log2(factor)
    return cent

def find_harm_interval(note,first_voice_sequence,scaleType,wav_mid,TpQ, docu ):
    '''
    Returns an array of correspoinding factors to first voice, that happens in the same time. 
    There could be a different number of event in the first and second voice, therefore ther is a collection of intervals
    '''
    low_mid=int(note['mid'])
    low_tick=int(note['tick'])                                                         #low_ refers to the low voice (which isnormally lower than the melody)
    low_length=int(note['length'])
    low_end=low_tick+low_length
    intervals=[]
    tempo=0
    for first in first_voice_sequence:                                                #Go throug all first voice midi

        if first["function"] =="Tempo":                                               #Read the tempo
            tempo=int(first["tempo"])   
            SecPerTick=tempo/(TpQ*1000000)
        if first["function"] =="note":                                                #first_ refers to the melody
            first_mid=int(first['mid'])                                                  
            first_tick=int(first['tick'])
            first_length=int(first['length'])  
                 
            first_end=first_tick+first_length
            if first_tick <low_tick and first_end>low_tick:                           #first voice starting before harmony but ending in the (harmony) note
                if first_end<low_end:
                    end=first_end
                else:
                    end=low_end 
                mid_diff= low_mid-first_mid
                factor=mid_2_freq_factor(mid_diff,scaleType)
                start=0
                length=first_end-low_tick       
                newInterval={"low_tick":low_tick, "start":start,"length":length, "end":end,"low_length":low_length,  "factor":factor,"mid":low_mid,"first_mid":first_mid,"first_tick":first_tick,"tempo":tempo,"SecPerTick":SecPerTick,"document":"first_tick_inside"   }
                intervals.append(newInterval)
                docu.append(newInterval)
            if(first_tick>=low_tick and first_tick<=low_end):                                   #finding where there is a firstnote starting at start or later than start the note
                if first_end<low_end:
                    end=first_end
                else:
                    end=low_end
                mid_diff= low_mid-first_mid                                                     #A negative difference giving a factor <0
                factor=mid_2_freq_factor(mid_diff,scaleType)    
                start=first_tick-low_tick
                
                length=first_tick-end
                newInterval={"low_tick":low_tick,"start":start,"length":length, "end":end,"low_length":low_length,  "factor":factor,"mid":low_mid,"first_mid":first_mid,"first_tick":first_tick,"tempo":tempo,"SecPerTick":SecPerTick,"document":"first_tick_outside"  }
                intervals.append(newInterval)
                docu.append(newInterval)
                LastTempo=tempo
            LastSecPerTick=SecPerTick

    if len(intervals)==0:                                                                       #If no first voice is found, then make the freuqency factor
        mid_diff=wav_mid-low_mid
        factor=mid_2_freq_factor(mid_diff,scaleType)
        start=0 
        length =low_end-low_tick
        newInterval={"low_tick":low_tick, "start":start,"length":length,  "end":low_end,"low_length":low_length, "factor":factor,"mid":low_mid,"first_mid":"","first_tick":"","tempo":LastTempo, "SecPerTick":LastSecPerTick,"document":"len_is_0" }   #use first_mid =false to sort out that the factor is set already to the 
        intervals.append(newInterval)
        docu.append(newInterval)

    #intervals,tick,tick_length,mid,tempo,key
    return intervals,low_tick,low_length,low_mid,tempo,LastSecPerTick,docu



def findFactors(first_sequ,intervals,tick,tick_length,tempo,tuning_factor,SeqPerTick,TpQ,wl,docu):
    '''
    Makes the arrays to be inserted into hpsTransformations_function based on the factors in first_sequ
    this is done by matching the first_Tick and the First_mid with the corresponding values in the first_sequ 
    And and multiplying the factor there. 

    '''
  
   
    for interval in intervals:
        low_tick=int(interval["low_tick"])        
        first_mid=int(interval["first_mid"])
        first_tick=int(interval["first_tick"])
        factor_low=float(interval['factor'])
        low_mid=int(interval['mid'])
        start=int(interval['start'])
        length=int(interval['length'])
        #SecPerTick=int(interval['SecPerTick']) 
        #[{'low_tick': 0, 'end': 455, 'factor': 1.3333333333333333, 'mid': 57, 'first_mid': 62, 'first_tick': 0, 'tempo': 500000}]
        i=0
        factor_pairs=""
        for first in first_sequ:                
            sequ_mid=int(first["mid"])                  #finding the correct factor in the sequence based on mid an tick information            
            sequ_tick=int(first["tick"])
            if first_mid==sequ_mid and first_tick==sequ_tick:
                first_factor=float(first['factor'])
                factor=factor_low*first_factor*tuning_factor
                mid_dif=low_mid-first_mid
                cent=factor_2_cent(factor)
                doc="factor:"+str(factor) +' first-mid:'+str(first_mid)+' mid:'+str(low_mid)+' dif:'+ str(mid_dif)+ ' cent: '+str(cent) 
                docu.append(doc)

                print(doc)

                if i==0:
                    factor_pairs+="0,"+str(factor)
                else:
                    factor_pairs+= ","+ str(start)+","+str(factor)    
        i+=1
    factors=(factor_pairs)
    # timeScaling=np.array([0,0,wl,lenSec])  
    lengths=[0,0,wl,length]



            
    return factors, docu




def convert_second_harmony(midi_sequence_harm,first_midi_sequence,first_sequ,scaleType,wav_mid,tuning_factor,TpQ,wl):
    '''
    The harmonies (second voice is converted with reference to to the melody)
    '''
    sequence=[]
    docu=[]
    tempo=0
    for note in midi_sequence_harm:
        if note["function"]=="note":
            vel=int(note["param1"])
            intervals,tick,length,mid,tempo,SecPerTick,docu=harm_Intervals=find_harm_interval(note,first_midi_sequence,scaleType,wav_mid,TpQ,docu)
            startSec=tick*SecPerTick
            lenSec=length*SecPerTick              
            factor_array, docu=findFactors(first_sequ,intervals,tick,length,tempo,tuning_factor,SecPerTick,TpQ,wl,docu)
            factor_array=literal_eval(factor_array)
            #factor_array=[factor_array]
            NewNote={"startSec":startSec,"lenSec":lenSec,"factor_array":factor_array,"mid":mid,"vel":vel,"tick":tick}
            docu.append(NewNote)
            print(str(NewNote))
            sequence.append(NewNote)
            
    return sequence,docu        


    

def convert_sequence(midi_sequence,TpQ,TrackLength,Tempos,scaletype="tempered",wav_mid=69,uselastnote=""): 
    '''
    This is the main convert sequence, but with the 
    '''    
    tick=0
    lastNote=find_last_note(midi_sequence,tick)                                     #The last note is used for finding the root note in Minor, Doric and Mixolydian melodies
                                                                                    #The scoring softwares do write down the number of # or b-s but they are not accurate around
                                                                                    #major or minor, which is important for this application.
    totTime=totalTime(Tempos,TpQ,TrackLength)                                       #seconds

    sequence=[]
    tempo=0
    docu=[]
    for note in midi_sequence:
        for temp in Tempos:
            te=int(temp['tempo'])
            ti=int(temp['tick'])
            if tick>ti:
                tempo=te


        if note["function"] =="Tempo":                                              #Read the tempo
            tempo=int(note["tempo"])
                                                                                     #microseconds per quarter note TpQ ticks per quarter note
        if note["function"] =="Key_signature":                                      #The key signature is not used yet, but could be in the future            
            key=int(note["param1"])  
            tick =int(note['tick'])                                                                               #Som Swedish and Norwegian music dialects vary the beat within each measure
            lastNote=find_last_note(midi_sequence,tick)    
            rootKey=midi_offset_due_2_key(key)
            if uselastnote:                                                        #setting uselastnote setting to any string evaluates true
                root=lastNote             
            else:
                root=rootKey                                                       #if you in the config.json sets uselastnote to "" you get the (Major) key setting from the midi file, The setting is irrelvant for 
            offs_from_wav=60+root-wav_mid
            base_factor=mid_2_freq_factor(offs_from_wav,scaletype)     
            rootchange={"root":root, "lastNote":lastNote,"rootKey":rootKey,"tick":tick}
            docu.append (rootchange) 
            #print("tonart:"+str(lastNote)+' '+str(offs_from_c) + ' '+ str(offs_key)+ ' ' +str(offs_from_wav)+' '+ str(base_factor) +' ' + uselastnote +str(tick) )      
        if note["function"]=="note":
            tick=int(note["tick"])
            mid=int(note["mid"])
            length=int(note["length"])
            vel=int(note["param1"])
            SecPerTick=tempo/(TpQ*1000000)
            lenSec=length*SecPerTick
            startSec=tick*SecPerTick
            middiff=mid-(60+root)
            pow2=middiff//12
            pow2_factor=pow(2, pow2)
            restmid=middiff-12*pow2
            rest_factor=mid_2_freq_factor(restmid,scaletype)
            factor=base_factor*pow2_factor*rest_factor
            event={"startSec":startSec,"lenSec":lenSec,"factor":factor,"tick":tick,"mid":mid,"vel":vel,"SecPerTick":SecPerTick}            
            sequence.append(event)
    return sequence,totTime,lastNote,docu  


def find_last_note(midi_sequence,starttick):
    #finding the last note value
    lastNote=0
    i=0    
    for note in midi_sequence:
        if int(note['tick'])>int(starttick):
            if note["function"]=="note":
                lastNote=int(note["mid"])    
            if note["function"]=="Key_signature":
                lastNote = lastNote % 12
                return lastNote                   
 
                
        i=i+1      
    lastNote = lastNote % 12
    return lastNote     
def midi_offset_due_2_key(key):
    #Key is number of #'s or number of b's (circle of fifth)  
    # value is midi-offset C =0 Db=1, D=2 Eb=3 etc reference is "keyhole C=60"
    # Normal music score programs does gives the key, but do not distinguish between minor and minor
    # There is an option of selecting the last note of the first voice which is normally the root note. 
    if key>6:
        key=key-12
    if  key<-6:
        key=key+12  
    key2mid_offs={
        -6:-6,            # Gb
        -5:1,             # Db
        -4:8,             # Ab
        -3:3,             # Eb
        -2:10,            # Bb
        -1:5,             # F
        0:0,              # C
        1:-5,             # G
        2:2,              # D
        3:9,              # A
        4:4,              # E
        5:11,             # B
        6:6               # F#
    }
    offs_from_c=key2mid_offs.get(key)  
    return offs_from_c



def mid_2_freq_factor(mid,scaleType):     
    if scaleType=="tempered":        
        factor=pow(2, mid/12)                                              #Freq calc with a tempered scale is simple
    else:
        scales={
            "pythagorean":pythagorean,
            "just":just, 
            "ptolomy_frygian_tc":frygian_tense_chromatic            
        }
        power=mid//12
        midDif=mid % 12
        if midDif<0:
            midDif=midDif+12
            power-=1
        scale=scales.get(scaleType)                                       #select the scale dictionary frome the scales dictionary
        T,N=scale.get(midDif)                                             # Get the Nominator (T) and Denominator (N) from the scale Dictionary
        factor=T/N*pow(2,power)                              
    return factor
    







            
        



    

