import py_midicsv as pm
import pandas as pd
import numpy as np
def endNote(tick,mid,stack,out,track):
    '''
    The first transformation while reading the notes, isto transform the two note on and note off events to one event
    Containing note on and note length. The velocity paramter is kept into the param1 
    '''
    idx=0
    
    for itm in stack:
        if itm["mid"]==mid:            
            length=int(tick)-int(itm["tick"])
            vel=itm["vel"]
            note={"function":"note","tick":itm["tick"],"mid":mid,"length":length,"param1":vel,"param2":0,"track":track}
            out.append(note)
            stack.pop(idx)
            break

        idx=idx+1    
    return out



def readmidi(fileName):
    '''
    Open the filename midifile and reads the file converts into an array with midi onset number and length
    of midi-pitch, and tick length  in other words converting on and off events to on events, with event length, which is what we need. 
    This module delivers pitch in Midi numbers and timings in ticks, they are converted to frequency factors and time in seconds in the converters.py
    '''
    Midi_out_Sequences=[]                                               #There may be several tracks in a midi file
    Midi_out_Sequence=[]                                                #Each track
    Key=0
    Tempos=[]                                                           #Intialize the Tempos array
    TrackLength=0
    Midi_in_Sequence = pm.midi_to_csv(fileName)                         #converts the binary Midi file to a text array
    Key_signatures=[]
    stack=[]
    Tempos=[]
    for event in Midi_in_Sequence:
        events =event.split(", ")        
        track=events[0]
        tick=events[1]
        function=events[2].rstrip()                                       #must strip \n char

        if function=="Header":
            
            if len(events)>4:           
                TpQ=int(events[5])                                        #ticks per Quarter note is written in the header

        if function=="Time_signature":
            nom=events[3]                                                 #Nominator 3 for waltz 2 or 4 for march etc. 
            denom=events[4].rstrip()                                      #Denominator id 2 to minus the power of denom 1=half note  2=quarternote 3=eigth note 4 sixtinth note etc.
            KeySign= {"function":"Time_signature","tick":tick,"mid":0,"length":0,"param1":nom,"param2":denom,"track":track}          
            Key_signatures.append(KeySign)
            Midi_out_Sequence.append(KeySign)
        if function=="Key_signature":                                     
            key=events[3]
            maj=events[4].rstrip()                                           # Key signatures may change, and are put back into the array so we can use it at the right time
            KeySign= {"function":"Key_signature","tick":tick,"mid":0,"length":0,"param1":key,"param2":maj,"track":track}          
            Key_signatures.append(KeySign)
            Midi_out_Sequence.append(KeySign)
        if function=='Tempo':
            tempo=events[3].rstrip()                                       #tempo is measured in microseconds per quaternote
            tempoChange={"function":"Tempo","tick":tick,"tempo":tempo}
            Tempos.append(tempoChange)                                     #With tempoChange variable we can easily calculate the total time if tempo has changed
            Midi_out_Sequence.append(tempoChange)


        if function=="End_track":                                          #when ending the track you get the global track length
            tick=events[1]                       
            TrackLength=int(tick)                                     


        if function=="Note_on_c":                                       #Getting the note start point
            track=events[0]
            tick=events[1]
            vel=int(events[5].rstrip())
            mid=events[4]
            

            if vel==0:
                endNote(tick,mid,stack,Midi_out_Sequence,track)          #There is two ways of ending anote, either set the velocity to zero or use Note_off_c
            else:
                noteval={"tick":tick,"mid":mid,"vel":vel}                #Append to stack if not "note off"
                stack.append(noteval)    
            
        if function=="End_track":
            Midi_out_Sequences.append(Midi_out_Sequence)
            Midi_out_Sequence=[]

        
        if function=="Note_off_c":                                        #Note off event
            track=events[0]
            tick=events[1]
            vel=events[5].rstrip()
            mid=events[4]            
            Midi_out_Sequence=endNote(tick,mid,stack,Midi_out_Sequence,track)  


    


    return Midi_out_Sequences, TrackLength, TpQ,Tempos


''' test
fileName="./mid/po-farmorsbrud.mid"
Midi_out_Sequence, TrackLength TpQ=readmidi(fileName)

print(TrackLength)
for note in Midi_out_Sequence:    
    print(note)
'''    